--==============================================================================
-- CERN (BE-CO-HT)
-- Xilinx MultiBoot core top-level file
--==============================================================================
--
-- author: Theodor Stana (t.stana@cern.ch)
--
-- date of creation: 2013-08-19
--
-- version: 1.0
--
-- description: 
--
-- dependencies:
--
-- references:
-- 
--==============================================================================
-- GNU LESSER GENERAL PUBLIC LICENSE
--==============================================================================
-- This source file is free software; you can redistribute it and/or modify it
-- under the terms of the GNU Lesser General Public License as published by the
-- Free Software Foundation; either version 2.1 of the License, or (at your
-- option) any later version. This source is distributed in the hope that it
-- will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- See the GNU Lesser General Public License for more details. You should have
-- received a copy of the GNU Lesser General Public License along with this
-- source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
--==============================================================================
-- last changes:
--    2013-08-19   Theodor Stana     t.stana@cern.ch     File created
--==============================================================================
-- TODO: - 
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library unisim;
use unisim.vcomponents.all;

use work.wishbone_pkg.all;

entity xil_multiboot is
  port
  (
    -- Clock and reset input ports
    clk_i   : in  std_logic;
    rst_n_i : in  std_logic;

    -- Wishbone ports
    wbs_i   : in  t_wishbone_slave_in;
    wbs_o   : out t_wishbone_slave_out;

    -- SPI ports
    spi_cs_n_o : out std_logic;
    spi_sclk_o : out std_logic;
    spi_mosi_o : out std_logic;
    spi_miso_i : in  std_logic
  );
end entity xil_multiboot;


architecture behav of xil_multiboot is

  --============================================================================
  -- Type declarations
  --============================================================================

  --============================================================================
  -- Constant declarations
  --============================================================================

  --============================================================================
  -- Component declarations
  --============================================================================
  -- Register component
  component multiboot_regs is
    port (
      -- Clock and reset inputs
      rst_n_i            : in     std_logic;
      clk_sys_i          : in     std_logic;

      -- Wishbone ports
      wb_adr_i           : in     std_logic_vector(2 downto 0);
      wb_dat_i           : in     std_logic_vector(31 downto 0);
      wb_dat_o           : out    std_logic_vector(31 downto 0);
      wb_cyc_i           : in     std_logic;
      wb_sel_i           : in     std_logic_vector(3 downto 0);
      wb_stb_i           : in     std_logic;
      wb_we_i            : in     std_logic;
      wb_ack_o           : out    std_logic;
      wb_stall_o         : out    std_logic;

      -- Fields of control register
      multiboot_cr_rdbootsts_o     : out    std_logic;
      multiboot_cr_iprog_o         : out    std_logic;
      --multiboot_cr_flr_o           : out    std_logic;
      --multiboot_cr_flw_o           : out    std_logic;

      -- Fields of status register
      multiboot_sr_bootsts_img_i   : in std_logic_vector(15 downto 0);
      multiboot_sr_valid_i         : in std_logic;
      --multiboot_sr_flrrdy_i        : in std_logic;
      --multiboot_sr_flwrdy_i        : in std_logic;

      -- Fields of bitstream address registers
      multiboot_gbbar_o            : out std_logic_vector(31 downto 0);
      multiboot_mbbar_o            : out std_logic_vector(31 downto 0);

      -- Fields of FAR register
      --multiboot_far_data_load_o    : out std_logic;
      multiboot_far_data_i         : in  std_logic_vector(23 downto 0);
      multiboot_far_data_o         : out std_logic_vector(23 downto 0);
      multiboot_far_nbytes_o       : out std_logic_vector(1 downto 0);
      multiboot_far_xfer_o         : out std_logic;
      multiboot_far_cs_o           : out std_logic;
      multiboot_far_ready_i        : in  std_logic

      ---- Fields of bitstream address registers
      --multiboot_flrdr_i             : in  std_logic_vector(31 downto 0);
      --multiboot_flwdr_o             : out std_logic_vector(31 downto 0)
    );
  end component multiboot_regs;

  -- FSM component
  component multiboot_fsm is
    port
    (
      -- Clock and reset inputs
      clk_i                : in  std_logic;
      rst_n_i              : in  std_logic;

      -- Control register inputs
      reg_rdbootsts_i      : in  std_logic;
      reg_wmb_i            : in  std_logic;
      reg_wgb_i            : in  std_logic;
      reg_iprog_i          : in  std_logic;
      --reg_flr_i            : in  std_logic;
      --reg_flw_i            : in  std_logic;

      -- Multiboot and golden bitstream start addresses
      reg_gbbar_i          : in  std_logic_vector(31 downto 0);
      reg_mbbar_i          : in  std_logic_vector(31 downto 0);

      -- Outputs to status register
      reg_bootsts_img_o    : out std_logic_vector(15 downto 0);
      reg_bootsts_valid_o  : out std_logic;

      -- Flash access register signals
      reg_far_data_i       : in  std_logic_vector(23 downto 0);
      reg_far_data_o       : out std_logic_vector(23 downto 0);
      reg_far_nbytes_i     : in  std_logic_vector(1 downto 0);
      reg_far_xfer_i       : in  std_logic;
      reg_far_cs_i         : in  std_logic;
      reg_far_ready_o      : out std_logic;

      -- SPI master signals
      spi_xfer_o           : out std_logic;
      spi_cs_o             : out std_logic;
      spi_data_i           : in  std_logic_vector(7 downto 0);
      spi_data_o           : out std_logic_vector(7 downto 0);
      spi_ready_i          : in  std_logic;
      ---- Flash read data and ready bit
      --reg_flrd_o           : out std_logic_vector(31 downto 0);
      --reg_flrrdy_o         : out std_logic;

      ---- Flash write data and ready bit
      --reg_flwd_i           : in  std_logic_vector(31 downto 0);
      --reg_flwrdy_o         : out std_logic;

      ---- Ports for the external flash controller component
      --fl_addr_o            : out std_logic_vector(23 downto 0);
      --fl_rstat_o           : out std_logic;
      --fl_serase_o          : out std_logic;
      --fl_write_o           : out std_logic;
      --fl_read_o            : out std_logic;
      --fl_endcmd_o          : out std_logic; 
      --fl_data_o            : out std_logic_vector(7 downto 0);
      --fl_data_i            : in  std_logic_vector(7 downto 0);
      --fl_ready_i           : in  std_logic;

      -- Ports for the external ICAP component
      icap_dat_i           : in  std_logic_vector(15 downto 0);
      icap_dat_o           : out std_logic_vector(15 downto 0);

      icap_busy_i          : in  std_logic;
      icap_ce_n_o          : out std_logic;
      icap_wr_n_o          : out std_logic
    );
  end component multiboot_fsm;

  component spi_master is
    generic(
      -- clock division ratio (SCLK = clk_sys_i / (2 ** g_div_ratio_log2).
      g_div_ratio_log2 : integer := 2;
      -- number of data bits per transfer
      g_num_data_bits  : integer := 2);
    port (
      clk_sys_i : in std_logic;
      rst_n_i   : in std_logic;
  
      -- state of the Chip select line (1 = CS active). External control
      -- allows for multi-transfer commands (SPI master itself does not
      -- control the state of spi_cs_n_o)
      cs_i : in std_logic;
  
      -- 1: start next transfer (using CPOL, DATA and SEL from the inputs below)
      start_i    : in  std_logic;
  
      -- Clock polarity: 1: slave clocks in the data on rising SCLK edge, 0: ...
      -- on falling SCLK edge
      cpol_i     : in  std_logic;
  
      -- TX Data input 
      data_i     : in  std_logic_vector(g_num_data_bits - 1 downto 0);
  
      -- 1: data_o contains the result of last read operation. Core is ready to initiate
      -- another transfer.
      ready_o    : out std_logic;
  
      -- data read from selected slave, valid when ready_o == 1.
      data_o     : out std_logic_vector(g_num_data_bits - 1 downto 0);
  
      -- these are obvious
      spi_cs_n_o : out std_logic;
      spi_sclk_o : out std_logic;
      spi_mosi_o : out std_logic;
      spi_miso_i : in  std_logic
      );
  end component spi_master;

--  -- Flash controller component
--  component m25p_flash is
--    port 
--    (
--      clk_sys_i : in std_logic;
--      rst_n_i   : in std_logic;
--
--      -- Wishbone registers (FAR register access)
--      far_data_load_i : in  std_logic;
--      far_data_i      : in  std_logic_vector(7 downto 0);
--      far_data_o      : out std_logic_vector(7 downto 0);
--      far_xfer_i      : in  std_logic;
--      far_ready_o     : out std_logic;
--      far_cs_i        : in  std_logic;
--
--      -- Data readout interface. 
--
--      -- 1: sets flash read address to addr_i
--      set_addr_i : in std_logic;
--
--      -- start address for read operations
--      addr_i : in std_logic_vector(23 downto 0);
--
--      -- data request: when 1, the controller reads subsequent bytes from
--      -- the flash, starting from addr_i address.
--      read_i : in std_logic;
--
--      -- read data output
--      data_o : out std_logic_vector(7 downto 0);
--
--      -- when 1, data_o contains a valid byte and the controller is ready to accept
--      -- another command
--      ready_o : out std_logic;
--
--      -- SPI bus, connect to the flash memory.
--      spi_cs_n_o : out std_logic;
--      spi_sclk_o : out std_logic;
--      spi_mosi_o : out std_logic;
--      spi_miso_i : in  std_logic
--    );
--  end component m25p_flash;

  --============================================================================
  -- Signal declarations
  --============================================================================
  -- Control and status register signals
  signal rdbootsts     : std_logic;
  signal iprog         : std_logic;
  signal flr           : std_logic;
  signal flw           : std_logic;
  signal bootsts_img   : std_logic_vector(15 downto 0);
  signal sr_valid      : std_logic;
  signal gbbar, mbbar  : std_logic_vector(31 downto 0);

  -- FSM signals
  signal fsm_icap_din  : std_logic_vector(15 downto 0);
  signal fsm_icap_dout : std_logic_vector(15 downto 0);

  -- Flash controller signals
  --signal far_data_load : std_logic;
  signal far_data_out  : std_logic_vector(23 downto 0);
  signal far_data_in   : std_logic_vector(23 downto 0);
  signal far_nbytes    : std_logic_vector(1 downto 0);
  signal far_xfer      : std_logic;
  signal far_cs        : std_logic;
  signal far_ready     : std_logic;

  -- SPI master signals
  signal spi_data_in   : std_logic_vector(7 downto 0);
  signal spi_data_out  : std_logic_vector(7 downto 0);
  signal spi_xfer      : std_logic;
  signal spi_cs        : std_logic;
  signal spi_ready     : std_logic;

  signal fl_addr       : std_logic_vector(23 downto 0);
  signal fl_rstat      : std_logic;
  signal fl_serase     : std_logic;
  signal fl_write      : std_logic;
  signal fl_read       : std_logic;
  signal fl_data_out   : std_logic_vector(7 downto 0);
  signal fl_data_in    : std_logic_vector(7 downto 0);
  signal fl_ready      : std_logic;
  signal fl_endcmd     : std_logic;
  signal flrd          : std_logic_vector(31 downto 0);
  signal flrrdy        : std_logic;
  signal flwd          : std_logic_vector(31 downto 0);
  signal flwrdy        : std_logic;

  -- ICAP signals
  signal icap_ce_n     : std_logic;
  signal icap_wr_n     : std_logic;
  signal icap_busy     : std_logic;
  signal icap_din      : std_logic_vector(15 downto 0);
  signal icap_dout     : std_logic_vector(15 downto 0);

--==============================================================================
--  architecture begin
--==============================================================================
begin

  --============================================================================
  -- Register component instantiation
  --============================================================================
  -- First, some unused signal assignments
  wbs_o.err <= '0';
  wbs_o.rty <= '0';

  -- Now, instantiate the component
  cmp_regs : multiboot_regs
    port map
    (
      rst_n_i                    => rst_n_i,
      clk_sys_i                  => clk_i,

      wb_adr_i                   => wbs_i.adr(4 downto 2),
      wb_dat_i                   => wbs_i.dat,
      wb_dat_o                   => wbs_o.dat,
      wb_cyc_i                   => wbs_i.cyc,
      wb_sel_i                   => wbs_i.sel,
      wb_stb_i                   => wbs_i.stb,
      wb_we_i                    => wbs_i.we,
      wb_ack_o                   => wbs_o.ack,
      wb_stall_o                 => wbs_o.stall,

      multiboot_cr_rdbootsts_o   => rdbootsts,
      multiboot_cr_iprog_o       => iprog,
      --multiboot_cr_flr_o         => flr,
      --multiboot_cr_flw_o         => flw,

      multiboot_sr_bootsts_img_i => bootsts_img,
      multiboot_sr_valid_i       => sr_valid,
      --multiboot_sr_flrrdy_i      => flrrdy,
      --multiboot_sr_flwrdy_i      => flwrdy,

      multiboot_gbbar_o          => gbbar,
      multiboot_mbbar_o          => mbbar,

      --multiboot_far_data_load_o  => far_data_load,
      multiboot_far_data_i       => far_data_in,
      multiboot_far_data_o       => far_data_out,
      multiboot_far_nbytes_o     => far_nbytes,
      multiboot_far_xfer_o       => far_xfer,
      multiboot_far_cs_o         => far_cs,
      multiboot_far_ready_i      => far_ready

      -- multiboot_flrdr_i          => flrd,
      -- multiboot_flwdr_o          => flwd
    );

  --============================================================================
  -- FSM component instantiation
  --============================================================================
  cmp_fsm : multiboot_fsm
    port map
    (
      clk_i                => clk_i,
      rst_n_i              => rst_n_i,

      reg_rdbootsts_i      => rdbootsts,
      reg_wmb_i            => '0', --wmb,
      reg_wgb_i            => '0', --wgb,
      reg_iprog_i          => iprog,
      --reg_flr_i            => flr,
      --reg_flw_i            => flw,

      reg_gbbar_i          => gbbar,
      reg_mbbar_i          => mbbar,

      reg_bootsts_img_o    => bootsts_img,
      reg_bootsts_valid_o  => sr_valid,

      reg_far_data_i       => far_data_out,
      reg_far_data_o       => far_data_in,
      reg_far_nbytes_i     => far_nbytes,
      reg_far_xfer_i       => far_xfer,
      reg_far_cs_i         => far_cs,
      reg_far_ready_o      => far_ready,

      spi_xfer_o           => spi_xfer,
      spi_cs_o             => spi_cs,
      spi_data_i           => spi_data_out,
      spi_data_o           => spi_data_in,
      spi_ready_i          => spi_ready,
      
      --fl_addr_o            => fl_addr,
      --fl_rstat_o           => fl_rstat,
      --fl_serase_o          => fl_serase,
      --fl_write_o           => fl_write,
      --fl_read_o            => fl_read,
      --fl_endcmd_o          => fl_endcmd,
      --fl_data_i            => fl_data_out,
      --fl_data_o            => fl_data_in,
      --fl_ready_i           => fl_ready,

      --reg_flrd_o           => flrd,
      --reg_flrrdy_o         => flrrdy,
      --reg_flwd_i           => flwd,
      --reg_flwrdy_o         => flwrdy,

      icap_dat_i           => fsm_icap_din,
      icap_dat_o           => fsm_icap_dout,
      icap_busy_i          => icap_busy,
      icap_ce_n_o          => icap_ce_n,
      icap_wr_n_o          => icap_wr_n
    );

  --============================================================================
  -- Flash controller instantiation
  --============================================================================
  cmp_spi_master : spi_master
    generic map
    (
      g_div_ratio_log2 => 0,
      g_num_data_bits  => 8
    )
    port map
    (
      clk_sys_i  => clk_i,
      rst_n_i    => rst_n_i,
      cs_i       => spi_cs,
      start_i    => spi_xfer,
      cpol_i     => '0',
      data_i     => spi_data_in,
      data_o     => spi_data_out,
      ready_o    => spi_ready,
      spi_cs_n_o => spi_cs_n_o,
      spi_sclk_o => spi_sclk_o,
      spi_mosi_o => spi_mosi_o,
      spi_miso_i => spi_miso_i
    );

  --============================================================================
  -- Xilinx ICAP logic
  --============================================================================
  -- First, bit-flip the data to/from the FSM
  icap_din( 0) <= fsm_icap_dout( 7);
  icap_din( 1) <= fsm_icap_dout( 6);
  icap_din( 2) <= fsm_icap_dout( 5);
  icap_din( 3) <= fsm_icap_dout( 4);
  icap_din( 4) <= fsm_icap_dout( 3);
  icap_din( 5) <= fsm_icap_dout( 2);
  icap_din( 6) <= fsm_icap_dout( 1);
  icap_din( 7) <= fsm_icap_dout( 0);
  icap_din( 8) <= fsm_icap_dout(15);
  icap_din( 9) <= fsm_icap_dout(14);
  icap_din(10) <= fsm_icap_dout(13);
  icap_din(11) <= fsm_icap_dout(12);
  icap_din(12) <= fsm_icap_dout(11);
  icap_din(13) <= fsm_icap_dout(10);
  icap_din(14) <= fsm_icap_dout( 9);
  icap_din(15) <= fsm_icap_dout( 8);

  fsm_icap_din( 0) <= icap_dout( 7);
  fsm_icap_din( 1) <= icap_dout( 6);
  fsm_icap_din( 2) <= icap_dout( 5);
  fsm_icap_din( 3) <= icap_dout( 4);
  fsm_icap_din( 4) <= icap_dout( 3);
  fsm_icap_din( 5) <= icap_dout( 2);
  fsm_icap_din( 6) <= icap_dout( 1);
  fsm_icap_din( 7) <= icap_dout( 0);
  fsm_icap_din( 8) <= icap_dout(15);
  fsm_icap_din( 9) <= icap_dout(14);
  fsm_icap_din(10) <= icap_dout(13);
  fsm_icap_din(11) <= icap_dout(12);
  fsm_icap_din(12) <= icap_dout(11);
  fsm_icap_din(13) <= icap_dout(10);
  fsm_icap_din(14) <= icap_dout( 9);
  fsm_icap_din(15) <= icap_dout( 8);

  -- and instantiate the ICAP component
  cmp_icap : ICAP_SPARTAN6
    port map
    (
      CLK   => clk_i,
      CE    => icap_ce_n,
      WRITE => icap_wr_n,
      I     => icap_din,
      O     => icap_dout,
      BUSY  => icap_busy
    );

end architecture behav;
--==============================================================================
--  architecture end
--==============================================================================
